@@ -144,7 +144,27 @@ int dbd_initialize(dbi_driver_t * driver)
      * this is called right after dbd_register_driver().
      * return -1 on error, 0 on success. if -1 is returned, the driver will not
      * be added to the list of available drivers. */
-    return 0;
+
+  /* this indicates the driver can be safely unloaded when libdbi is
+     shut down. Change the value to '0' (zero) if the driver, or a
+     library it is linked against, installs exit handlers via
+     atexit() */
+  _dbd_register_driver_cap(driver, "safe_dlclose", 1);
+
+  /* this indicates the database engine supports transactions */
+  _dbd_register_driver_cap(driver, "transaction_support", 1);
+
+  /* this indicates the database engine supports savepoints */
+  _dbd_register_driver_cap(driver, "savepoint_support", 1);
+  return 0;
+}
+
+int dbd_finalize(dbi_driver_t *driver) {
+	/* perform any database-specific client library shutdown.
+	 * this is called right before dlclose()ing the driver.
+	 * return -1 on error, 0 on success. */
+
+	return 0;
 }
 
 int dbd_connect(dbi_conn_t * conn)
@@ -294,9 +314,10 @@ int dbd_free_query(dbi_result_t * result)
     return 0;
 }
 
-int dbd_goto_row(dbi_result_t * result, unsigned long long row)
+int dbd_goto_row(dbi_result_t * result, unsigned long long rowidx, unsigned long long currowidx)
 {
-    return 1;
+  /* no-op */
+  return 1;
 }
 
 int dbd_get_socket(dbi_conn_t * conn)
@@ -715,6 +736,76 @@ dbi_result_t *dbd_query_null(dbi_conn_t * conn, const unsigned char *statement,
     return NULL;
 }
 
+int dbd_transaction_begin(dbi_conn_t *conn) {
+  if (dbd_query(conn, "BEGIN TRANSACTION") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_commit(dbi_conn_t *conn) {
+  if (dbd_query(conn, "COMMIT TRANSACTION") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_rollback(dbi_conn_t *conn) {
+  if (dbd_query(conn, "ROLLBACK TRANSACTION") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "SAVE TRANSACTION %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_rollback_to_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "ROLLBACK TRANSACTION %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_release_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  /* this is apparently not used by SYBASE/SQL Server */
+        return 0;
+}
+
 const char *dbd_select_db(dbi_conn_t * conn, const char *db)
 {
     dbi_result_t *res;
@@ -732,10 +823,10 @@ const char *dbd_select_db(dbi_conn_t * conn, const char *db)
     return (char *) db;
 }
 
-int dbd_geterror(dbi_conn_t * conn, int *errno, char **errstr)
+int dbd_geterror(dbi_conn_t * conn, int *err_no, char **errstr)
 {
     /* put error number into errno, error string into errstr
-     * return 0 if error, 1 if errno filled, 2 if errstr filled, 3 if both errno and errstr filled */
+     * return 0 if error, 1 if err_no filled, 2 if errstr filled, 3 if both err_no and errstr filled */
 
     /* We havn't functions for read error types in freetds */
     return -1;
@@ -824,6 +915,7 @@ void _translate_freetds_type(CS_DATAFMT * datafmt, unsigned short *type, unsigne
     case CS_VARCHAR_TYPE:
 	_type = DBI_TYPE_STRING;
 	datafmt->format = CS_FMT_NULLTERM;
+	++datafmt->maxlength; /* 1 extra byte for \0 */
 	break;
     case CS_MONEY_TYPE:	/* 8 */
     case CS_MONEY4_TYPE:	/* 4 */
@@ -977,8 +1069,9 @@ dbi_row_t *_dbd_freetds_buffers_binding(dbi_conn_t * conn, dbi_result_t * result
 	    /* 
 	     * Result is more that 8 bytes - 
 	     * allocate additional memory
+	     * 1 extra byte for \0
 	     */
-	    addr = row->field_values[idx].d_string = (char *) malloc(row->field_sizes[idx]);
+	    addr = row->field_values[idx].d_string = (char *) malloc(row->field_sizes[idx] + 1);
 	    break;
 	default:
 	    /* Prepare union to data copy */
