@@ -17,7 +17,7 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
- * $Id: dbd_ingres.c,v 1.23 2006/06/12 16:58:17 qu1j0t3 Exp $
+ * $Id: dbd_ingres.c,v 1.28 2013/01/09 21:30:19 mhoenicka Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -69,14 +69,16 @@ static II_PTR envHandle = NULL;
 
 #define AUTOCOMMIT_ON(conn)		(((ingres_conn_t*)conn->connection)->autocommit)
 
-#define SAVE_ERROR(conn,errhdl)	ingres_error(errhdl, dbi_verbosity>1, \
+static dbi_inst_t_pointer dbi_instance = NULL;
+
+#define SAVE_ERROR(conn,errhdl)	ingres_error(errhdl, dbi_instance->dbi_verbosity>1, \
 									 &((ingres_conn_t*)conn->connection)->errorCode, \
 									 &((ingres_conn_t*)conn->connection)->errorMsg)
-#define DEBUG_ERROR(errhdl)		ingres_error(errhdl, dbi_verbosity>2, NULL, NULL)
+#define DEBUG_ERROR(errhdl)		ingres_error(errhdl, dbi_instance->dbi_verbosity>2, NULL, NULL)
 #define DRIVER_ERROR(conn,msg)	_dbd_internal_error_handler(conn,msg,0)
 
-#define PRINT_VERBOSE	if(dbi_verbosity>1) _verbose_handler
-#define PRINT_DEBUG		if(dbi_verbosity>2) _verbose_handler
+#define PRINT_VERBOSE	if(dbi_instance->dbi_verbosity>1) _verbose_handler
+#define PRINT_DEBUG		if(dbi_instance->dbi_verbosity>2) _verbose_handler
 
 #define IS_BLOB(T) ( (T) == IIAPI_LVCH_TYPE  || (T) == IIAPI_LBYTE_TYPE \
 				  || (T) == IIAPI_LNVCH_TYPE || (T) == IIAPI_LTXT_TYPE )
@@ -117,6 +119,19 @@ int dbd_initialize(dbi_driver_t *driver) {
 	 * this is called right after dbd_register_driver().
 	 * return -1 on error, 0 on success. if -1 is returned, the driver will not
 	 * be added to the list of available drivers. */
+
+        /* this indicates the driver can be safely unloaded when libdbi is
+	   shut down. Change the value to '0' (zero) if the driver, or a
+	   library it is linked against, installs exit handlers via
+	   atexit() */
+        _dbd_register_driver_cap(driver, "safe_dlclose", 1);
+
+	/* this indicates the database engine supports transactions */
+        _dbd_register_driver_cap(driver, "transaction_support", 1);
+
+	/* this indicates the database engine supports savepoints */
+        _dbd_register_driver_cap(driver, "savepoint_support", 1);
+
 	IIAPI_INITPARM  initParm;
 
 	initParm.in_version = IIAPI_VERSION_4;
@@ -124,6 +139,7 @@ int dbd_initialize(dbi_driver_t *driver) {
 	IIapi_initialize( &initParm );
 	if(initParm.in_status < IIAPI_ST_ERROR){
 		envHandle = initParm.in_envHandle;
+		dbi_instance = driver->dbi_inst;
 		return 0;
 	}else{
 		fputs("failed to initialise Ingres driver\n", stderr);
@@ -132,6 +148,15 @@ int dbd_initialize(dbi_driver_t *driver) {
 	// FIXME: need to call terminate at driver shutdown.
 }
 
+int dbd_finalize(dbi_driver_t *driver) {
+	/* perform any database-specific client library shutdown.
+	 * this is called right before dlclose()ing the driver.
+	 * return -1 on error, 0 on success. */
+
+  /* todo: see above FIXME */
+	return 0;
+}
+
 
 /* ---------- API helpers ---------- */
 
@@ -386,11 +411,11 @@ int dbd_ping(dbi_conn_t *conn) {
 	return test;
 }
 
-int dbd_geterror(dbi_conn_t *conn, int *errno, char **errstr) {
-	/* put error number into errno, error string into errstr
-	 * return 0 if error, 1 if errno filled, 2 if errstr filled, 3 if both errno and errstr filled */
+int dbd_geterror(dbi_conn_t *conn, int *err_no, char **errstr) {
+	/* put error number into err_no, error string into errstr
+	 * return 0 if error, 1 if err_no filled, 2 if errstr filled, 3 if both err_no and errstr filled */
 	if(conn && conn->connection && ((ingres_conn_t*)conn->connection)->errorMsg){
-		*errno = ((ingres_conn_t*)conn->connection)->errorCode;
+		*err_no = ((ingres_conn_t*)conn->connection)->errorCode;
 		*errstr = strdup( ((ingres_conn_t*)conn->connection)->errorMsg );
 		return 3;
 	}
@@ -852,8 +877,9 @@ int dbd_free_query(dbi_result_t *result) {
 	return 0;
 }
 
-int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx) {
-	return 1;
+int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx, unsigned long long currowidx) {
+  /* no-op */
+  return 1;
 }
 
 
@@ -1004,7 +1030,74 @@ dbi_result_t *dbd_query_null(dbi_conn_t *conn, const unsigned char *statement, s
 	return NULL;
 }
 
+int dbd_transaction_begin(dbi_conn_t *conn) {
+  /* transactions begin implicitly after CONNECT, COMMIT, or ROLLBACK statements */
+  return 0;
+}
+
+int dbd_transaction_commit(dbi_conn_t *conn) {
+  if (dbd_query(conn, "COMMIT") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_rollback(dbi_conn_t *conn) {
+  if (dbd_query(conn, "ROLLBACK") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_rollback_to_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "ROLLBACK TO %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_release_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  /* this is apparently not used by Ingres */
+        return 0;
+}
+
 static dbi_result_t *ingres_sys_query(dbi_conn_t *conn, const char *sql) {
+        dbi_inst instance;
 	dbi_result_t *res = NULL;
 	ingres_conn_t *iconn = conn->connection;
 	
@@ -1012,7 +1105,9 @@ static dbi_result_t *ingres_sys_query(dbi_conn_t *conn, const char *sql) {
 		// note that we need to keep the system connection around,
 		// because closing it prematurely will kill result sets
 		// FIXME: could use _disjoin_from_conn() instead
-		iconn->sysConn = dbi_conn_new(driver_info.name);
+
+                instance = dbi_driver_get_instance(dbi_conn_get_driver(conn));
+		iconn->sysConn = dbi_conn_new_r(driver_info.name, instance);
 		if(ingres_connect(iconn->sysConn, SYS_CATALOGS, NO_AUTOCOMMIT) < 0){
 			_verbose_handler(conn,"can't connect to '%s'\n",SYS_CATALOGS);
 			return NULL;
@@ -1051,11 +1146,15 @@ dbi_result_t *dbd_list_dbs(dbi_conn_t *conn, const char *pattern) {
 }
 
 dbi_result_t *dbd_list_tables(dbi_conn_t *conn, const char *db, const char *pattern) {
+        dbi_inst instance;
 	static char *select = "SELECT table_name FROM iitables WHERE table_name NOT LIKE 'ii%'";
 	char *sql = select;
 	dbi_result_t *res = NULL;
-	dbi_conn_t *newconn = dbi_conn_new(driver_info.name);
-	
+	dbi_conn_t *newconn;
+
+        instance = dbi_driver_get_instance(dbi_conn_get_driver(conn));
+        newconn = dbi_conn_new_r(driver_info.name, instance);
+
 	if(ingres_connect(newconn, db, NO_AUTOCOMMIT) < 0)
 		_verbose_handler(conn,"dbd_list_tables: can't connect to '%s'\n",db);
 	else{
