@@ -21,7 +21,7 @@
  * Copyright (C) 2001-2002, Mark Tobenkin <mark@brentwoodradio.com>
  * http://libdbi.sourceforge.net
  * 
- * $Id: dbd_mysql.c,v 1.95 2008/01/15 00:27:29 mhoenicka Exp $
+ * $Id: dbd_mysql.c,v 1.109 2013/02/27 23:16:20 mhoenicka Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -51,7 +51,7 @@ long long strtoll(const char *nptr, char **endptr, int base);
 #include <dbi/dbi-dev.h>
 #include <dbi/dbd.h>
 
-#include <mysql/mysql.h>
+#include <mysql.h>
 #include "dbd_mysql.h"
 
 static const dbi_info_t driver_info = {
@@ -63,7 +63,7 @@ static const dbi_info_t driver_info = {
 	__DATE__
 };
 
-static const char *custom_functions[] = {NULL}; // TODO
+static const char *custom_functions[] = MYSQL_CUSTOM_FUNCTIONS;
 static const char *reserved_words[] = MYSQL_RESERVED_WORDS;
 
 /* encoding strings, array is terminated by a pair of empty strings */
@@ -124,23 +124,65 @@ void dbd_register_driver(const dbi_info_t **_driver_info, const char ***_custom_
 int dbd_initialize(dbi_driver_t *driver) {
 	/* perform any database-specific server initialization.
 	 * this is called right after dbd_register_driver().
-	 * return -1 on error, 0 on success. if -1 is returned, the driver will not
+	 * return -1 on error, 0 on success.
+	 * if -1 is returned, the driver will not
 	 * be added to the list of available drivers. */
+
+        /* this indicates the driver can be safely unloaded when libdbi is
+	   shut down. Change the value to '0' (zero) if the driver, or a
+	   library it is linked against, installs exit handlers via
+	   atexit() */
         _dbd_register_driver_cap(driver, "safe_dlclose", 1);
+
+	/* this indicates the database engine supports transactions */
+        _dbd_register_driver_cap(driver, "transaction_support", 1);
+
+	/* this indicates the database engine supports savepoints */
+        _dbd_register_driver_cap(driver, "savepoint_support", 1);
+	return 0;
+}
+
+int dbd_finalize(dbi_driver_t *driver) {
+	/* perform any database-specific client library shutdown.
+	 * this is called right before dlclose()ing the driver.
+	 * return -1 on error, 0 on success. */
+
+        /* free memory allocated by the library. mysql_server_end() is
+	   deprecated as of MySQL 5.0.3, but the replacement
+	   mysql_library_end() is simply an alias defined in
+	   mysql.h. Leave this as is until it breaks */
+        mysql_server_end();
 	return 0;
 }
 
 int dbd_connect(dbi_conn_t *conn) {
 	MYSQL *mycon;
+	dbi_result_t *result = NULL;
 	char* sql_cmd;
 	unsigned long client_flags = 0;
+	long n_port = 0;
 
 	const char *host = dbi_conn_get_option(conn, "host");
 	const char *username = dbi_conn_get_option(conn, "username");
 	const char *password = dbi_conn_get_option(conn, "password");
 	const char *dbname = dbi_conn_get_option(conn, "dbname");
 	const char *encoding = dbi_conn_get_option(conn, "encoding");
-	int port = dbi_conn_get_option_numeric(conn, "port");
+	const char *port = dbi_conn_get_option(conn, "port");
+	if (port) {
+	  n_port = strtol(port, NULL, 10);
+	}
+	else {
+	  n_port = (long)dbi_conn_get_option_numeric(conn, "port");
+	}
+
+	/* if no port was specified, use MySQL's default port to let
+	   things run gracefully */
+	if (n_port == 0) {
+	  n_port = 3306;
+	}
+
+	int timeout = dbi_conn_get_option_numeric(conn, "timeout");
+
 	/* mysql specific options */
 	const char *unix_socket = dbi_conn_get_option(conn, "mysql_unix_socket");
 	/* we honor the mysql_compression flag for the sake of
@@ -162,7 +204,14 @@ int dbd_connect(dbi_conn_t *conn) {
 		_dbd_internal_error_handler(conn, NULL, DBI_ERROR_NOMEM);
 		return -2;
 	}
-	else if (!mysql_real_connect(mycon, host, username, password, dbname, port, unix_socket, client_flags)) {
+
+	if (timeout != -1) { /* option was specified */
+	  /* the mysql_options prototype is braindead */
+	  mysql_options(mycon, MYSQL_OPT_CONNECT_TIMEOUT, (const char*) &timeout);
+	}
+
+	if (!mysql_real_connect(mycon, host, username, password, dbname, (unsigned int)n_port, unix_socket, client_flags)) {
+/* 	  printf("mysql_real_connect failed with host=%s\nusername=%s\npassword=%s\ndbname=%s\nport=%s\n", username,password,dbname,port); */
 		conn->connection = (void *)mycon; // still need this set so _error_handler can grab information
 		_dbd_internal_error_handler(conn, NULL, DBI_ERROR_DBD);
 		mysql_close(mycon);
@@ -173,33 +222,25 @@ int dbd_connect(dbi_conn_t *conn) {
 		conn->connection = (void *)mycon;
 		if (dbname) conn->current_db = strdup(dbname);
 	}
-/* 	printf("dbname went to %s\n", dbname);	 */
+
 	if (encoding && *encoding) {
 	  /* set connection encoding */
 	  if (!strcmp(encoding, "auto")) {
 	    encoding = dbd_get_encoding(conn);
 	    if (encoding) {
 	      asprintf(&sql_cmd, "SET NAMES '%s'", dbd_encoding_from_iana(encoding));
-/*  	      printf("SET NAMES '%s'\n", dbd_encoding_from_iana(encoding)); */
-	      dbd_query(conn, sql_cmd);
+	      result = dbd_query(conn, sql_cmd);
 	      free(sql_cmd);
+	      dbi_result_free(result);
 	    }
-/* 	    else { */
-/* 	      printf("could not retrieve encoding\n"); */
-/* 	    } */
-	    /* else: do nothing, use default */
 	  }
 	  else {
 	    asprintf(&sql_cmd, "SET NAMES '%s'", dbd_encoding_from_iana(encoding));
-/*   	    printf("SET NAMES '%s'", dbd_encoding_from_iana(encoding)); */
-	    dbd_query(conn, sql_cmd);
+	    result = dbd_query(conn, sql_cmd);
 	    free(sql_cmd);
+	    dbi_result_free(result);
 	  }
-/*  	  printf("set encoding to %s<<\n", dbd_encoding_from_iana(encoding)); */
 	}
-/* 	else { */
-/* 	  printf("leave encoding alone\n"); */
-/* 	} */
 
 	return 0;
 }
@@ -229,12 +270,14 @@ int dbd_free_query(dbi_result_t *result) {
 	return 0;
 }
 
-int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx) {
-	// XXX TODO: kosherize this, handle efficient queries.
+int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx, unsigned long long currowidx) {
+  if (rowidx != currowidx+1) {
 	mysql_data_seek((MYSQL_RES *)result->result_handle, rowidx);
 	/* the return type of this function is indeed void, so it is
 	   unclear what happens if rowidx is outside the range. The
 	   calling function must make sure the row index is valid */
+  }
+  /* else: nothing to do, next fetch will fetch next row */
 	return 1;
 }
 
@@ -506,7 +549,15 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
 	
 	res = mysql_store_result((MYSQL *)conn->connection);
 	
-	/* if res is null, the query was something that doesn't return rows (like an INSERT) */
+	/* if res is null and mysql reports an error,
+	   we encountered a select failure */
+	if (!res && mysql_errno((MYSQL *)conn->connection)) {
+	  return NULL;
+	}
+
+	/* else: if res is null and no error is indicated, the
+	   query was something that doesn't return rows (like an
+	   INSERT) */
 	result = _dbd_result_create(conn, (void *)res, (res ? mysql_num_rows(res) : 0), 
 								mysql_affected_rows((MYSQL *)conn->connection));
 
@@ -514,6 +565,9 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
 	  _dbd_result_set_numfields(result, mysql_num_fields((MYSQL_RES *)result->result_handle));
 	  _get_field_info(result);
 	}
+	else {
+	  _dbd_result_set_numfields(result, 0);
+	}
 
 	return result;
 }
@@ -540,6 +594,90 @@ dbi_result_t *dbd_query_null(dbi_conn_t *conn, const unsigned char *statement, s
 	return result;
 }
 
+int dbd_transaction_begin(dbi_conn_t *conn) {
+  if (dbd_query(conn, "START TRANSACTION") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_commit(dbi_conn_t *conn) {
+  if (dbd_query(conn, "COMMIT") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_rollback(dbi_conn_t *conn) {
+  if (dbd_query(conn, "ROLLBACK") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_rollback_to_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "ROLLBACK TO SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_release_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "RELEASE SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
 const char *dbd_select_db(dbi_conn_t *conn, const char *db) {
 	if (mysql_select_db((MYSQL *)conn->connection, db)) {
 		_dbd_internal_error_handler(conn, NULL, DBI_ERROR_DBD);
@@ -553,15 +691,15 @@ const char *dbd_select_db(dbi_conn_t *conn, const char *db) {
 	return db;
 }
 
-int dbd_geterror(dbi_conn_t *conn, int *errno, char **errstr) {
-	/* put error number into errno, error string into errstr
-	 * return 0 if error, 1 if errno filled, 2 if errstr filled, 3 if both errno and errstr filled */
+int dbd_geterror(dbi_conn_t *conn, int *err_no, char **errstr) {
+	/* put error number into err_no, error string into errstr
+	 * return 0 if error, 1 if err_no filled, 2 if errstr filled, 3 if both err_no and errstr filled */
 
 	if (strcmp("",mysql_error((MYSQL *)conn->connection)) == 0) {
 		return -1;
 	}
 	
-	*errno = mysql_errno((MYSQL *)conn->connection);
+	*err_no = mysql_errno((MYSQL *)conn->connection);
 	*errstr = strdup(mysql_error((MYSQL *)conn->connection));
 	return 3;
 }
@@ -642,8 +780,6 @@ void _translate_mysql_type(MYSQL_FIELD *field, unsigned short *type, unsigned in
 			_attribs |= DBI_DATETIME_DATE | DBI_DATETIME_TIME;
 			break;
 			
-		case FIELD_TYPE_VAR_STRING:
-		case FIELD_TYPE_STRING:
 		case FIELD_TYPE_TINY_BLOB:
 		case FIELD_TYPE_MEDIUM_BLOB:
 		case FIELD_TYPE_LONG_BLOB:
@@ -655,6 +791,8 @@ void _translate_mysql_type(MYSQL_FIELD *field, unsigned short *type, unsigned in
 				_type = DBI_TYPE_BINARY;
 				break;
 			}
+		case FIELD_TYPE_VAR_STRING:
+		case FIELD_TYPE_STRING:
 #ifdef FIELD_TYPE_NEWDECIMAL
 		case FIELD_TYPE_NEWDECIMAL: // Precision math DECIMAL or NUMERIC field (MySQL 5.0.3 and up)
 #endif
