@@ -21,7 +21,7 @@
  * Copyright (C) 2001-2002, David A. Parker <david@neongoat.com>.
  * http://libdbi.sourceforge.net
  * 
- * $Id: dbd_pgsql.c,v 1.55 2008/01/26 17:39:50 mhoenicka Exp $
+ * $Id: dbd_pgsql.c,v 1.69 2013/02/08 01:01:31 mhoenicka Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -65,7 +65,7 @@ static const dbi_info_t driver_info = {
 	__DATE__
 };
 
-static const char *custom_functions[] = {NULL}; // TODO
+static const char *custom_functions[] = PGSQL_CUSTOM_FUNCTIONS;
 static const char *reserved_words[] = PGSQL_RESERVED_WORDS;
 
 /* encoding strings, array is terminated by a pair of empty strings */
@@ -75,6 +75,7 @@ static const char pgsql_encoding_hash[][16] = {
   "EUC_JP", "EUC-JP",
   "EUC_KR", "EUC-KR",
   "UNICODE", "UTF-8",
+  "UTF8", "UTF-8",
   "LATIN1", "ISO-8859-1",
   "LATIN2", "ISO-8859-2",
   "LATIN3", "ISO-8859-3",
@@ -100,6 +101,8 @@ void _translate_postgresql_type(unsigned int oid, unsigned short *type, unsigned
 void _get_field_info(dbi_result_t *result);
 void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned long long rowidx);
 int _dbd_real_connect(dbi_conn_t *conn, const char *db);
+char *_unescape_hex_binary(char* raw, size_t in_len, size_t* out_len);
+int _digit_to_number(const char c);
 
 /* this function is available through the PostgreSQL client library, but it
    is not declared in any of their headers. I hope this won't break anything */
@@ -127,6 +130,26 @@ const char *pg_encoding_to_char(int encoding_id);
             asprintf( &conninfo, fmt, key, value );                  \
 	} while(0)
 
+
+/* base36 decoding to convert the 5 alphanumeric chars of SQLSTATE to a 32bit int */
+int base36decode(char *base36) {
+    int len = strlen(base36);
+    int output = 0;
+    int pos = 0;
+
+    for (; pos < len; pos++) {
+        char c = base36[pos];
+        if ( ((c - '0') >= 0) && ((c - '0') <= 9) ) {
+            c = c - '0';
+        } else {
+            c = c - 'A' + 10;
+        }
+        output = 36 * output + c;
+    }
+
+    return output;
+}
+
 /* real code starts here */
 void dbd_register_driver(const dbi_info_t **_driver_info, const char ***_custom_functions, const char ***_reserved_words) {
 	/* this is the first function called after the driver module is loaded into memory */
@@ -141,7 +164,26 @@ int dbd_initialize(dbi_driver_t *driver) {
 	 * return -1 on error, 0 on success. if -1 is returned, the driver will not
 	 * be added to the list of available drivers. */
 	
+        /* this indicates the driver can be safely unloaded when libdbi is
+	   shut down. Change the value to '0' (zero) if the driver, or a
+	   library it is linked against, installs exit handlers via
+	   atexit() */
         _dbd_register_driver_cap(driver, "safe_dlclose", 1);
+
+	/* this indicates the database engine supports transactions */
+        _dbd_register_driver_cap(driver, "transaction_support", 1);
+
+	/* this indicates the database engine supports savepoints */
+        _dbd_register_driver_cap(driver, "savepoint_support", 1);
+
+	return 0;
+}
+
+int dbd_finalize(dbi_driver_t *driver) {
+	/* perform any database-specific client library shutdown.
+	 * this is called right before dlclose()ing the driver.
+	 * return -1 on error, 0 on success. */
+
 	return 0;
 }
 
@@ -160,6 +202,7 @@ int _dbd_real_connect(dbi_conn_t *conn, const char *db) {
 	const char *pgopt;
 	const char *optval;
 	int         optval_num;
+	int have_port = 0;
 
 	/* PQconnectdb accepts additional options as a string of
 	   "key=value" pairs. Assemble that string from the option
@@ -176,6 +219,10 @@ int _dbd_real_connect(dbi_conn_t *conn, const char *db) {
 	    pgopt = "user";
 	  }
 
+	  else if (!strcmp(pgopt, "timeout")) {
+	    pgopt = "connect_timeout";
+	  }
+
 	  /* Map "pgsql_foo" to "foo" */
 	  else if( !strncmp( pgopt, "pgsql_", 6 ) ) {
 	    pgopt += 6;
@@ -188,6 +235,10 @@ int _dbd_real_connect(dbi_conn_t *conn, const char *db) {
 	    continue;
 	  }
 
+	  if (!strcmp(pgopt, "port")) {
+	    have_port++;
+	  }
+
 	  optval     = dbi_conn_get_option( conn, optname );
 	  optval_num = dbi_conn_get_option_numeric( conn, optname );
 
@@ -209,6 +260,11 @@ int _dbd_real_connect(dbi_conn_t *conn, const char *db) {
 	if( dbname )
 		CONNINFO_APPEND_ESCAPED( conninfo, "%s='%s'", "dbname", dbname );
 
+	/* if no port was specified, fill in the default PostgreSQL port */
+	if (!have_port) {
+	    CONNINFO_APPEND( conninfo, "%s='%d'", "port", 5432 );
+	}
+	  
 	/* send an empty string instead of NULL if there are no options */
 	pgconn = PQconnectdb(conninfo ? conninfo : "");
 	if (conninfo) free(conninfo);
@@ -241,8 +297,10 @@ int _dbd_real_connect(dbi_conn_t *conn, const char *db) {
 }
 
 int dbd_disconnect(dbi_conn_t *conn) {
-	if (conn->connection) PQfinish((PGconn *)conn->connection);
-	return 0;
+  if (conn->connection) {
+    PQfinish((PGconn *)conn->connection);
+  }
+  return 0;
 }
 
 int dbd_fetch_row(dbi_result_t *result, unsigned long long rowidx) {
@@ -265,7 +323,7 @@ int dbd_free_query(dbi_result_t *result) {
 	return 0;
 }
 
-int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx) {
+int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx, unsigned long long currowidx) {
 	/* libpq doesn't have to do anything, the row index is specified when
 	 * fetching fields */
 	return 1;
@@ -360,48 +418,7 @@ const char* dbd_encoding_from_iana(const char *iana_encoding) {
 }
 
 char *dbd_get_engine_version(dbi_conn_t *conn, char *versionstring) {
-  dbi_result_t *dbi_result;
-  const char *versioninfo = NULL;
-
-  /* initialize return string */
-  *versionstring = '\0';
-
-  dbi_result = dbd_query(conn, "SELECT VERSION()");
-
-  /* this query will return something like:
-     PostgreSQL 8.0.1 on i386-portbld-freebsd5.4, compiled by GCC cc (GCC) 3.4.2 [FreeBSD] 20040728
-  */
-  if (dbi_result) {
-    if (dbi_result_next_row(dbi_result)) {
-      char *dot = NULL;
-      char *start = NULL;
-      char *stop = NULL;
-      versioninfo = dbi_result_get_string_idx(dbi_result, 1);
-
-      /* try to locate the version number. Look for the first dot, go
-	 back where the number before the dot starts, then walk
-	 forward to the last dot or number */
-      dot = strchr(versioninfo, (int)'.');
-      if (dot) {
-	start = dot-1;
-	while (start>versioninfo && isdigit((int)(*(start-1)))) {
-	  start--;
-	}
-
-	stop = start;
-	while (*(stop+1) && (isdigit((int)(*(stop+1))) || *(stop+1)=='.')) {
-	  stop++;
-	}
-
-	if (stop-start < VERSIONSTRING_LENGTH) {
-	  strncpy(versionstring, start, stop-start+1);
-	  versionstring[stop-start+1] = '\0';
-	}
-      }
-    }
-    dbi_result_free(dbi_result);
-  }
-
+  snprintf(versionstring, VERSIONSTRING_LENGTH, "%d", PQserverVersion((PGconn *)conn->connection));
   return versionstring;
 }
 
@@ -453,7 +470,7 @@ size_t dbd_quote_binary(dbi_conn_t *conn, const unsigned char* orig, size_t from
   unsigned char *quoted_temp = NULL;
   size_t to_length;
 
-  temp = PQescapeBytea(orig, from_length, &to_length);
+  temp = PQescapeByteaConn((PGconn *)conn->connection, orig, from_length, &to_length);
 
   if (!temp) {
     return 0;
@@ -471,7 +488,9 @@ size_t dbd_quote_binary(dbi_conn_t *conn, const unsigned char* orig, size_t from
   PQfreemem((void*)temp);
 
   *ptr_dest = quoted_temp;
-  return to_length;
+
+  /* to_length already contains one extra byte for the trailing NULL byte */
+  return to_length+1;
 }
 
 dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
@@ -486,11 +505,15 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
 	
 	res = PQexec((PGconn *)conn->connection, statement);
 	if (res) resstatus = PQresultStatus(res);
-	if (!res || ((resstatus != PGRES_COMMAND_OK) && (resstatus != PGRES_TUPLES_OK))) {
-		PQclear(res);
-		return NULL;
+	if (!res || ((resstatus != PGRES_COMMAND_OK) && (resstatus != PGRES_TUPLES_OK) && (resstatus != PGRES_COPY_OUT) && (resstatus != PGRES_COPY_IN))) {
+	  char *base36 = PQresultErrorField(res, PG_DIAG_SQLSTATE);
+	  conn->error_number = (! base36) ? 0 : base36decode(base36);
+	  PQclear(res);
+	  return NULL;
 	}
 
+	conn->error_number = 0;
+
 	result = _dbd_result_create(conn, (void *)res, (unsigned long long)PQntuples(res), (unsigned long long)atoll(PQcmdTuples(res)));
 	_dbd_result_set_numfields(result, (unsigned int)PQnfields((PGresult *)result->result_handle));
 	_get_field_info(result);
@@ -502,6 +525,90 @@ dbi_result_t *dbd_query_null(dbi_conn_t *conn, const unsigned char *statement, s
 	return NULL;
 }
 
+int dbd_transaction_begin(dbi_conn_t *conn) {
+  if (dbd_query(conn, "BEGIN TRANSACTION") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_commit(dbi_conn_t *conn) {
+  if (dbd_query(conn, "COMMIT") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_rollback(dbi_conn_t *conn) {
+  if (dbd_query(conn, "ROLLBACK") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_rollback_to_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "ROLLBACK TO SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_release_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "RELEASE SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
 const char *dbd_select_db(dbi_conn_t *conn, const char *db) {
   /* postgresql doesn't support switching databases without reconnecting */
   if (!db || !*db) {
@@ -520,14 +627,14 @@ const char *dbd_select_db(dbi_conn_t *conn, const char *db) {
   return db;
 }
 
-int dbd_geterror(dbi_conn_t *conn, int *errno, char **errstr) {
-	/* put error number into errno, error string into errstr
-	 * return 0 if error, 1 if errno filled, 2 if errstr filled, 3 if both errno and errstr filled */
+int dbd_geterror(dbi_conn_t *conn, int *err_no, char **errstr) {
+	/* put error number into err_no, error string into errstr
+	 * return 0 if error, 1 if err_no filled, 2 if errstr filled, 3 if both err_no and errstr filled */
 	
-	*errno = 0;
+	*err_no = conn->error_number;
 	*errstr = strdup(PQerrorMessage((PGconn *)conn->connection));
 	
-	return 2;
+	return 3;
 }
 
 unsigned long long dbd_get_seq_last(dbi_conn_t *conn, const char *sequence) {
@@ -744,21 +851,34 @@ void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned long long rowi
 				data->d_string = strdup(raw);
 				row->field_sizes[curfield] = strsize;
 				break;
-			case DBI_TYPE_BINARY:	
-			  temp = PQunescapeBytea((const unsigned char *)raw, &unquoted_length);
-			  if ((data->d_string = malloc(unquoted_length)) == NULL) {
+			case DBI_TYPE_BINARY:
+			  strsize = (size_t)PQgetlength((PGresult *)result->result_handle, rowidx, curfield);
+			  if (strsize > 2
+			      && raw[0] == '\\'
+			      && raw[1] == 'x') {
+			    /* hex format */
+				/* row->field_sizes[curfield] = strsize; */
+				/* data->d_string = malloc(strsize); */
+				/* memcpy(data->d_string, raw, strsize); */
+			    temp = PQunescapeBytea((const unsigned char *)_unescape_hex_binary(raw, strsize, &unquoted_length), &(row->field_sizes[curfield]));
+			    if ((data->d_string = malloc(row->field_sizes[curfield])) == NULL) {
+			      PQfreemem(temp);
+			      break;
+			    }
+			    memmove(data->d_string, temp, row->field_sizes[curfield]);
 			    PQfreemem(temp);
-			    break;
 			  }
-			  memmove(data->d_string, temp, unquoted_length);
-			  PQfreemem(temp);
-			  row->field_sizes[curfield] = unquoted_length;
-			  /* todo: is raw ever unescaped binary data? */
-/*        			        strsize = PQgetlength((PGresult *)result->result_handle, rowidx, curfield); */
-/* 				row->field_sizes[curfield] = strsize; */
-/* 				data->d_string = malloc(strsize); */
-/* 				memcpy(data->d_string, raw, strsize); */
-				break;
+			  else {
+			    temp = PQunescapeBytea((const unsigned char *)raw, &unquoted_length);
+			    if ((data->d_string = malloc(unquoted_length)) == NULL) {
+			      PQfreemem(temp);
+			      break;
+			    }
+			    memmove(data->d_string, temp, unquoted_length);
+			    PQfreemem(temp);
+			    row->field_sizes[curfield] = unquoted_length;
+			  }
+			  break;
 				
 			case DBI_TYPE_DATETIME:
 				sizeattrib = result->field_attribs[curfield] & (DBI_DATETIME_DATE|DBI_DATETIME_TIME);
@@ -773,3 +893,81 @@ void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned long long rowi
 	}
 }
 
+/* this function reverts the changes done by PQescapeByteaConn to a
+   binary string. libpq does not provide such a function.  Returns the
+   result as a malloc'ed string which must be freed by the caller. The
+   output string is in the BYTEA escape format with single backslashes
+   and single single quotes. It must be post-processed by
+   PQunescapeBytea() to obtain true binary data
+*/
+char *_unescape_hex_binary(char* raw, size_t in_len, size_t* out_len) {
+  size_t i;
+  int in_pair = 0;
+  int last_nibble = 0;
+  char *outstring;
+  char *end_of_outstring;
+  int have_backslash = 0;
+  int have_singlequote = 0;
+  char tempchar;
+
+  /* algorithm borrowed and modified from:
+     http://pqxx.org/development/libpqxx/browser/trunk/src/binarystring.cxx
+  */
+
+  if ((outstring = malloc(((in_len-2)/2)+1)) == NULL) {
+    return NULL;
+  }
+  end_of_outstring = outstring;
+
+  for (i=2; i<in_len; ++i) {
+    const unsigned char c = raw[i];
+    if (isspace(c)) {
+      if (in_pair) {
+	/* "Escaped binary data is malformed." */
+      }
+    }
+    else if (!isxdigit(c)) {
+      /* "Escaped binary data contains invalid characters." */
+    }
+    else {
+      const int nibble = (isdigit(c) ? _digit_to_number(c) : (10 + tolower(c) - 'a'));
+      if (in_pair) {
+	tempchar = (char)((last_nibble<<4) | nibble);
+	if (tempchar == '\\' && have_backslash) {
+	  /* skip second consecutive backslash */
+	  have_backslash = 0;
+	}
+	else if (tempchar == '\'' && have_singlequote) {
+	  /* skip second consecutive single quote */
+	  have_singlequote = 0;
+	}
+	else {
+	  if (tempchar == '\\') {
+	    have_backslash = 1;
+	  }
+	  else if (tempchar == '\'') {
+	    have_singlequote = 1;
+	  }
+	  else {
+	    have_backslash = 0;
+	    have_singlequote = 0;
+	  }
+	  *end_of_outstring = tempchar;
+	  end_of_outstring++;
+	}
+      }
+      else {
+	last_nibble = nibble;
+      }
+      in_pair = !in_pair;
+    }
+  }
+  *end_of_outstring = '\0';
+  *out_len = end_of_outstring-outstring;
+  return outstring;
+}
+
+/* converts an ASCII character code to an integer */
+int _digit_to_number(const char c) {
+  return (int)c - (int)'0';
+}
