@@ -22,7 +22,7 @@
  * Copyright (C) 2002-2007, Markus Hoenicka <mhoenicka@users.sourceforge.net>
  * http://libdbi-drivers.sourceforge.net
  * 
- * $Id: dbd_sqlite.c,v 1.39 2008/01/15 14:07:24 mhoenicka Exp $
+ * $Id: dbd_sqlite.c,v 1.55 2013/01/09 21:30:20 mhoenicka Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -31,6 +31,9 @@
 
 #define _GNU_SOURCE /* we need asprintf */
 
+/* this is defined by the Makefile and passed via -D */
+/* #define DBDIR /usr/local/var/lib/libdbi/sqlite */
+
 #ifndef HAVE_ATOLL
 long long atoll(const char *str);
 #endif
@@ -66,8 +69,9 @@ static const dbi_info_t driver_info = {
   __DATE__
 };
 
-static const char *custom_functions[] = {NULL}; // TODO
+static const char *custom_functions[] = SQLITE_CUSTOM_FUNCTIONS;
 static const char *reserved_words[] = SQLITE_RESERVED_WORDS;
+static const char default_dbdir[] = DBDIR;
 
 /* the encoding strings */
 static const char sqlite_encoding_UTF8[] = "UTF-8";
@@ -84,7 +88,7 @@ static size_t sqlite_escape_string(char *to, const char *from, size_t length);
 int wild_case_compare(const char *str,const char *str_end,
 		      const char *wildstr,const char *wildend,
 		      char escape);
-
+static const char* _conn_get_dbdir(dbi_conn_t *conn);
 
 /* the real functions */
 void dbd_register_driver(const dbi_info_t **_driver_info, const char ***_custom_functions, const char ***_reserved_words) {
@@ -100,11 +104,28 @@ int dbd_initialize(dbi_driver_t *driver) {
    * return -1 on error, 0 on success. if -1 is returned, the driver will not
    * be added to the list of available drivers. */
 	
+  /* this indicates the driver can be safely unloaded when libdbi is
+     shut down. Change the value to '0' (zero) if the driver, or a
+     library it is linked against, installs exit handlers via
+     atexit() */
   _dbd_register_driver_cap(driver, "safe_dlclose", 1);
+
+  /* this indicates the database engine supports transactions */
+  _dbd_register_driver_cap(driver, "transaction_support", 1);
+
+  /* this indicates the database engine supports savepoints */
+  _dbd_register_driver_cap(driver, "savepoint_support", 1);
+
   return 0;
 }
 
+int dbd_finalize(dbi_driver_t *driver) {
+	/* perform any database-specific client library shutdown.
+	 * this is called right before dlclose()ing the driver.
+	 * return -1 on error, 0 on success. */
 
+	return 0;
+}
 
 int dbd_connect(dbi_conn_t *conn) {
   /* connect using the database set with the "dbname" option */
@@ -145,7 +166,7 @@ int _real_dbd_connect(dbi_conn_t *conn, const char* database) {
   }
 
   /* sqlite specific options */
-  dbdir = dbi_conn_get_option(conn, "sqlite_dbdir");
+  dbdir = _conn_get_dbdir(conn);
 	
   if (!dbdir) {
     _dbd_internal_error_handler(conn, "no database directory specified", DBI_ERROR_CLIENT);
@@ -199,11 +220,17 @@ int _real_dbd_connect(dbi_conn_t *conn, const char* database) {
     }
   }
 
-  /* set the SQLite timeout to timeout milliseconds */
+  /* set the SQLite timeout to timeout milliseconds. The older
+     SQLite-specific setting takes precedence over the generic timeout
+     option for backwards compatibility */
   timeout = dbi_conn_get_option_numeric(conn, "sqlite_timeout");
 
   if (timeout == -1) {
-    timeout = 0;
+    /* generic timeout is specified in seconds, not milliseconds */
+    timeout = 1000*dbi_conn_get_option_numeric(conn, "timeout");
+    if (timeout == -1) {
+      timeout = 0;
+    }
   }
 
   sqlite_busy_timeout(sqcon, timeout);	
@@ -247,7 +274,7 @@ int dbd_free_query(dbi_result_t *result) {
   return 0;
 }
 
-int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx) {
+int dbd_goto_row(dbi_result_t *result, unsigned long long rowidx, unsigned long long currowidx) {
   result->currowidx = rowidx;
   return 1;
 }
@@ -304,15 +331,18 @@ dbi_result_t *dbd_list_dbs(dbi_conn_t *conn, const char *pattern) {
   char *sq_errmsg = NULL;
   char old_cwd[_POSIX_PATH_MAX] = "";
   int retval;
+  size_t entry_size;
   DIR *dp;
   struct dirent *entry;
+  struct dirent *result;
   struct stat statbuf;
+  dbi_result rs;
 
   /* sqlite has no builtin function to list databases. Databases are just
      files in the data directory. We search for matching files and fill a
      temporary table with what we've found. Then we query this table and
      pretend sqlite has done all the work */
-  const char *sq_datadir = dbi_conn_get_option(conn, "sqlite_dbdir");
+  const char *sq_datadir = _conn_get_dbdir(conn);
 
   /* this is not nice but we have to drop the table even if it does not
    exist (sqlite has no way to list *temporary* tables so we can't check
@@ -321,17 +351,39 @@ dbi_result_t *dbd_list_dbs(dbi_conn_t *conn, const char *pattern) {
    Update: Now apparently there is a system table that lists
    temporary tables, but the DROP TABLE error doesn't hurt and is
    most likely faster than checking for the existence of the table */
-  dbd_query(conn, "DROP TABLE libdbi_databases");
-  dbd_query(conn, "CREATE TEMPORARY TABLE libdbi_databases (dbname VARCHAR(255))");
+  rs = dbd_query(conn, "DROP TABLE libdbi_databases");
+  dbi_result_free(rs);
+  rs = dbd_query(conn, "CREATE TEMPORARY TABLE libdbi_databases (dbname VARCHAR(255))");
+  dbi_result_free(rs);
 
   if (sq_datadir && (dp = opendir(sq_datadir)) == NULL) {
     _dbd_internal_error_handler(conn, "could not open data directory", DBI_ERROR_CLIENT);
     return NULL;
   }
+
+  /* allocate memory for readdir_r(3) */
+  entry_size = _dirent_buf_size(dp);
+  if (entry_size == 0) {
+    return NULL;
+  }
+
+  entry = (struct dirent *) malloc (entry_size);
+  if (entry == NULL) {
+    return NULL;
+  }
+
+  memset (entry, 0, entry_size);
+
   getcwd(old_cwd, _POSIX_PATH_MAX);
   chdir(sq_datadir);
 
-  while ((entry = readdir(dp)) != NULL) {
+  while (1) {
+    result = NULL;
+    retval = readdir_r(dp, entry, &result);
+    if (retval != 0 || result == NULL) {
+      break;
+    }
+
     stat(entry->d_name, &statbuf);
     if (S_ISREG(statbuf.st_mode)) {
       /* we do a magic number check here to make sure we
@@ -388,7 +440,9 @@ dbi_result_t *dbd_list_dbs(dbi_conn_t *conn, const char *pattern) {
       }
       /* else: we can't read it, so forget about it */
     }
-  }
+  } /* end while */
+
+  free(entry);
   closedir(dp);
   chdir(old_cwd);
 
@@ -403,28 +457,34 @@ dbi_result_t *dbd_list_tables(dbi_conn_t *conn, const char *db, const char *patt
    */
   dbi_result_t *dbi_result;
   dbi_conn_t* tempconn;
+  dbi_inst instance;
   int retval;
-  char* sq_errmsg;
+  char* sq_errmsg = NULL;
   char* sql_cmd;
+  dbi_result_t *rs;
 
   /* this function tries to query a specific database, so we need a
    separate connection to that other database, retrieve the table names,
    and feed them to a temporary table in our main connection */
-  tempconn = dbi_conn_new("sqlite");
+  instance = dbi_driver_get_instance(dbi_conn_get_driver(conn));
+  tempconn = dbi_conn_new_r("sqlite", instance);
 
   /* we explicitly cast to (char*) as we discard the "const" thing here */
   dbi_conn_set_option(tempconn, "dbname", (char*)db);
-  dbi_conn_set_option(tempconn, "sqlite_dbdir", (char*)dbi_conn_get_option(conn, "sqlite_dbdir"));
+  dbi_conn_set_option(tempconn, "sqlite_dbdir", (char*)_conn_get_dbdir(conn));
 
   if (dbi_conn_connect(tempconn) < 0) {
     _dbd_internal_error_handler(conn, NULL, DBI_ERROR_NOCONN);
+    dbi_conn_close(tempconn);
     return NULL;
   }
   
   /* create temporary table for table names. The DROP command won't hurt
      if the table doesn't exist yet */
-  dbd_query(conn, "DROP TABLE libdbi_tablenames");
-  dbd_query(conn, "CREATE TEMPORARY TABLE libdbi_tablenames (tablename VARCHAR(255))");
+  rs = dbd_query(conn, "DROP TABLE libdbi_tablenames");
+  dbi_result_free(rs);
+  rs = dbd_query(conn, "CREATE TEMPORARY TABLE libdbi_tablenames (tablename VARCHAR(255))");
+  dbi_result_free(rs);
 /*   fprintf(stderr, "created temporary table\n"); */
 
   /* sqlite does not support the SHOW command, so we have to extract the
@@ -440,16 +500,16 @@ dbi_result_t *dbd_list_tables(dbi_conn_t *conn, const char *db, const char *patt
 /*   fprintf(stderr, "select from sqlite_master has run\n"); */
   if (dbi_result) {
     while (dbi_result_next_row(dbi_result)) {
-      retval = sqlite_exec_printf((sqlite*)(conn->connection), "INSERT INTO libdbi_tablenames VALUES ('%s')", NULL, NULL, &sq_errmsg, dbi_result_get_string(dbi_result, "name"));
+      retval = sqlite_exec_printf((sqlite*)(conn->connection), "INSERT INTO libdbi_tablenames VALUES ('%s')", NULL, NULL, NULL /* no error messages please */, dbi_result_get_string(dbi_result, "name"));
     }
     dbi_result_free(dbi_result);
   }
   else {
     dbi_conn_error(tempconn, (const char**)&sq_errmsg);
-    free(sq_errmsg);
   }
 
-  sqlite_close((sqlite*)(tempconn->connection));
+  /* sqlite_close((sqlite*)(tempconn->connection)); */
+  dbi_conn_close(tempconn);
 
   return dbd_query(conn, "SELECT tablename FROM libdbi_tablenames ORDER BY tablename");
 }
@@ -521,6 +581,9 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
 
   if (query_res) {
     _dbd_internal_error_handler(conn, errmsg, query_res);
+    if (result_table != NULL) {
+      sqlite_free_table(result_table);
+    }
     return NULL;
   }
 	
@@ -554,6 +617,90 @@ dbi_result_t *dbd_query(dbi_conn_t *conn, const char *statement) {
   return result;
 }
 
+int dbd_transaction_begin(dbi_conn_t *conn) {
+  if (dbd_query(conn, "BEGIN") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_commit(dbi_conn_t *conn) {
+  if (dbd_query(conn, "COMMIT") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_transaction_rollback(dbi_conn_t *conn) {
+  if (dbd_query(conn, "ROLLBACK") == NULL) {
+    return 1;
+  }
+  else {
+    return 0;
+  }
+}
+
+int dbd_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_rollback_to_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "ROLLBACK TO SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
+int dbd_release_savepoint(dbi_conn_t *conn, const char *savepoint) {
+  char* query;
+
+  if (!savepoint) {
+    return 1;
+  }
+
+  asprintf(&query, "RELEASE SAVEPOINT %s", savepoint);
+
+  if (dbd_query(conn, query) == NULL) {
+    free(query);
+    return 1;
+  }
+  else {
+    free(query);
+    return 0;
+  }
+}
+
 int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement) {
   /* 
      field is the name of the field which we want to know the type of
@@ -580,6 +727,7 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
 
   char* item;
   char* table;
+  char* my_statement = NULL;
   char curr_table[MAX_IDENT_LENGTH] = "";
   char curr_field_name[MAX_IDENT_LENGTH];
   char curr_field_name_up[MAX_IDENT_LENGTH];
@@ -596,52 +744,93 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
    notation "table.field" is used */
   item = strchr(field, (int)'.');
   if (!item) {
-    /* the fields do not contain the table info. This means that
-       all fields are from the same table which we have to extract
-       from the statement that created the result */
+    /* the field does not contain the table info. However, the latter
+     may be available in the original statement, so let's look
+     there first*/
+    my_statement = strdup(statement);
+    if (!my_statement) {
+      return 0;
+    }
 
-    /* To get started, we use the first item after 'from' or 'FROM'
-       as the table name (we currently ignore pathologic cases like
-       'FroM' or 'froM'. We could uppercase a copy but we need the
-       table name as is, so it is going to get complex) */
-    if (!(table = strstr(statement, " from "))) {
-      table = strstr(statement, " FROM ");
+    if (!(table = strstr(my_statement, " from "))) {
+      table = strstr(my_statement, " FROM ");
     }
 
     if (!table) {
 /*       fprintf(stderr, "no from keyword found\n"); */
       return 0;
     }
-    
-    /* set ptr to possible start of item after 'from' */
-    table += 6;
 
-    /* skip spaces */
-    while (*table == ' ') {
-      table++;
-    }
+    *table = '\0'; /* terminate string, leaves only field names */
 
-    /* table now points to the table name; find the end of table */
-    item = table;
-    while (*item && *item != ' ' && *item != ',' && *item != ';') {
-      item++;
-    }
-    strncpy(curr_table, table, item-table);
-    curr_table[item-table] = '\0'; /* terminate just in case */
+    if ((table = strstr(my_statement, field)) != NULL
+	&& table != my_statement
+	&& *(table-1) == '.') {
+      /* the field name is there, isolate preceding table */
+      *(table-1) = '\0';
 
-    /* for obvious reasons, the internal tables do not contain the
-       commands how they were created themselves. We have to use known
-       values for the field types */
-    if (!strcmp(curr_table, "sqlite_master") ||
-	!strcmp(curr_table, "sqlite_temp_master")) {
-      if (!strcmp(field, "rootpage")) {
-	return FIELD_TYPE_LONG;
+      while (table > my_statement
+	     && *table != ' '
+	     && *table != ',') {
+	table--;
       }
-      else {
-	return FIELD_TYPE_STRING;
+
+      if (*table == ' '
+	  || *table == ',') {
+	table++;
+      }
+
+      /* table should now point to the table name */
+      strcpy(curr_table, table);
+    }
+    else {
+      /* as a last resort assume that all fields are from the same table
+	 which we have to extract from the statement that created the
+	 result */
+
+      /* To get started, we use the first item after 'from' or 'FROM'
+	 as the table name (we currently ignore pathologic cases like
+	 'FroM' or 'froM'. We could uppercase a copy but we need the
+	 table name as is, so it is going to get complex) */
+      if (!(table = strstr(statement, " from "))) {
+	table = strstr(statement, " FROM ");
+      }
+      
+      if (!table) {
+	/*       fprintf(stderr, "no from keyword found\n"); */
+	return 0;
+      }
+      
+      /* set ptr to possible start of item after 'from' */
+      table += 6;
+      
+      /* skip spaces */
+      while (*table == ' ') {
+	table++;
+      }
+
+      /* table now points to the table name; find the end of table */
+      item = table;
+      while (*item && *item != ' ' && *item != ',' && *item != ';') {
+	item++;
+      }
+      strncpy(curr_table, table, item-table);
+      curr_table[item-table] = '\0'; /* terminate just in case */
+
+      /* for obvious reasons, the internal tables do not contain the
+	 commands how they were created themselves. We have to use known
+	 values for the field types */
+      if (!strcmp(curr_table, "sqlite_master") ||
+	  !strcmp(curr_table, "sqlite_temp_master")) {
+	if (!strcmp(field, "rootpage")) {
+	  return FIELD_TYPE_LONG;
+	}
+	else {
+	  return FIELD_TYPE_STRING;
+	}
       }
     }
-
+    free(my_statement);
     strcpy(curr_field_name, field);
   }
   else {   /* each field contains table info */
@@ -704,6 +893,9 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
 				      curr_table);
 
   if (query_res || !table_numrows) {
+	if(table_result_table != NULL) {
+	  sqlite_free_table(table_result_table);
+	}
     /* now try in the table ocntaining temporary tables */
     query_res = sqlite_get_table_printf((sqlite*)conn->connection,
 					"SELECT tbl_name, sql FROM sqlite_temp_master where tbl_name='%s'",
@@ -715,6 +907,9 @@ int find_result_field_types(char* field, dbi_conn_t *conn, const char* statement
     
     if (query_res || !table_numrows) {
       _dbd_internal_error_handler(conn, errmsg, query_res);
+      if (table_result_table != NULL) {
+	sqlite_free_table(table_result_table);
+      }
 /*       printf("field not found\n"); */
       return 0;
     }
@@ -829,6 +1024,7 @@ char* get_field_type(const char* statement, const char* curr_field_name) {
     if an error occurred
   */
   char *item;
+  char* saveptr;
   char *my_statement;
   char *field_name;
   char *end_field_name;
@@ -852,7 +1048,7 @@ char* get_field_type(const char* statement, const char* curr_field_name) {
   item++;
 
   /* now tokenize the field list */
-  for (item = strtok(item, ","); item; item = strtok(NULL, ",")) {
+  for (item = strtok_r(item, ",", &saveptr); item; item = strtok_r(NULL, ",", &saveptr)) {
 /*     printf("item:%s<<\n", item); */
 
     /* skip leading whitespace */
@@ -911,13 +1107,13 @@ const char *dbd_select_db(dbi_conn_t *conn, const char *db) {
   return db;
 }
 
-int dbd_geterror(dbi_conn_t *conn, int *errno, char **errstr) {
-  /* put error number into errno, error string into errstr
-   * return 0 if error, 1 if errno filled, 2 if errstr filled, 3 if both errno and errstr filled */
+int dbd_geterror(dbi_conn_t *conn, int *err_no, char **errstr) {
+  /* put error number into err_no, error string into errstr
+   * return 0 if error, 1 if err_no filled, 2 if errstr filled, 3 if both err_no and errstr filled */
   int result = 0;
 
   if (conn->error_number) {
-    *errno = conn->error_number;
+    *err_no = conn->error_number;
     result++;
   }
   if (conn->error_message) {
@@ -1044,6 +1240,7 @@ void _get_row_data(dbi_result_t *result, dbi_row_t *row, unsigned int rowidx) {
     /* this will be set to the string size later on if the field is indeed a string */
 
     if (raw == NULL) { /* no data available */
+      _set_field_flag( row, curfield, DBI_VALUE_NULL, 1);
       curfield++;
       continue;
     }
@@ -1219,3 +1416,17 @@ static size_t sqlite_escape_string(char *to, const char *from, size_t length)
   return (size_t) (to-to_start);
 }
 
+/* this is a convenience function to retrieve the database directory */
+static const char* _conn_get_dbdir(dbi_conn_t *conn) {
+  const char* dbdir;
+
+  dbdir = dbi_conn_get_option(conn, "sqlite_dbdir");
+	
+  if (!dbdir) {
+    /* use default directory instead */
+    dbdir = default_dbdir;
+  }
+
+  return dbdir;
+}
+
